from __future__ import annotations
from typing import Any

import pickle
import redis
import time

from .base import _BaseCache, _MISSING


class RedisCache(_BaseCache):
    """Persistent cache backed by Redis."""

    def __init__(
        self,
        url: str | None = None,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password: str | None = None,
        namespace: str = "picocache",
        ttl: int | None = None,
        **kw: Any,
    ) -> None:
        super().__init__(**kw)
        self._r = (
            redis.Redis.from_url(url)
            if url
            else redis.Redis(host=host, port=port, db=db, password=password)
        )
        self._ns = namespace + ":"
        self._default_ttl = ttl
        self._lru_key = self._ns + "lru_tracker"  # Key for the sorted set tracking LRU

    def _lookup(self, key: str):
        full_key = self._ns + key
        data = self._r.get(full_key)
        if data is None:
            return _MISSING
        # Update LRU score on hit
        self._r.zadd(self._lru_key, {full_key: time.time()})
        self._hits += 1
        try:
            value = pickle.loads(data)
            return value
        except Exception as e:
            # Log error appropriately in a real app
            return _MISSING

    def _store(self, key: str, value: Any, wrapper_maxsize: int | None = None):
        full_key = self._ns + key
        try:
            pickled = pickle.dumps(value, protocol=self._PROTO)
            # Use a pipeline for atomicity
            pipe = self._r.pipeline()
            if self._default_ttl is None:
                pipe.set(full_key, pickled)
            else:
                pipe.setex(full_key, self._default_ttl, pickled)
            # Add/update key in LRU tracker only if maxsize is set for the wrapper
            if wrapper_maxsize is not None and wrapper_maxsize > 0:
                pipe.zadd(self._lru_key, {full_key: time.time()})
            pipe.execute()

            # Eviction logic is now handled by _evict_if_needed

        except Exception as e:
            # Log error appropriately in a real app
            pass  # Optionally raise

    def _evict_if_needed(self, wrapper_maxsize: int | None = None):
        # Eviction logic using wrapper_maxsize
        if wrapper_maxsize is not None and wrapper_maxsize > 0:
            # Check size *after* the new item has potentially been added by _store
            current_size = self._r.zcard(self._lru_key)
            if current_size > wrapper_maxsize:
                # Number of items to remove
                num_to_remove = current_size - wrapper_maxsize
                # Get the keys of the oldest items (lowest scores)
                # We want the oldest, so range from 0 up to num_to_remove - 1
                keys_to_remove = self._r.zrange(self._lru_key, 0, num_to_remove - 1)
                if keys_to_remove:
                    # Use pipeline to remove data keys and LRU entries
                    evict_pipe = self._r.pipeline()
                    evict_pipe.delete(*keys_to_remove)
                    evict_pipe.zrem(self._lru_key, *keys_to_remove)
                    evict_pipe.execute()

    def _clear(self) -> None:
        """Clear all keys within the namespace using SCAN."""
        cursor = 0
        match = self._ns + "*"
        while True:
            cursor, keys = self._r.scan(cursor=cursor, match=match)
            if keys:
                self._r.delete(*keys)
            if cursor == 0:
                break
        # Also delete the LRU tracker key *after* the loop
        self._r.delete(self._lru_key)

    def _get_current_size(self) -> int:
        """Return the number of keys tracked by the LRU mechanism."""
        # Always return the size of the LRU tracker set.
        # If maxsize wasn't used, the set won't be populated by _store.
        return self._r.zcard(self._lru_key)



from __future__ import annotations
from typing import Any

import pickle
import sqlite3
import threading
import time

from .base import _BaseCache, _MISSING


class SQLiteCache(_BaseCache):
    """Persistent cache backed by SQLite."""

    _TABLE_NAME = "picocache"

    def __init__(
        self,
        db_path: str = "picocache.db",
        **kw: Any,
    ) -> None:
        super().__init__(**kw)
        self._db_path = db_path
        self._lock = threading.Lock()
        self._conn = self._init_db()

    def _init_db(self) -> sqlite3.Connection:
        conn = sqlite3.connect(self._db_path, check_same_thread=False)
        with conn:
            conn.execute(
                f"""CREATE TABLE IF NOT EXISTS {self._TABLE_NAME} (
                    key TEXT PRIMARY KEY,
                    value BLOB,
                    last_accessed REAL
                )"""
            )
        return conn

    def _lookup(self, key: str) -> Any | _MISSING:
        with self._lock:
            cursor = self._conn.cursor()
            try:
                cursor.execute(
                    f"SELECT value FROM {self._TABLE_NAME} WHERE key = ?", (key,)
                )
                row = cursor.fetchone()
                if row is None:
                    return _MISSING

                # Update last_accessed time on lookup
                cursor.execute(
                    f"UPDATE {self._TABLE_NAME} SET last_accessed = ? WHERE key = ?",
                    (time.time(), key),
                )
                self._conn.commit()
                self._hits += 1  # Increment hit counter
                return pickle.loads(row[0])
            finally:
                cursor.close()

    def _store(self, key: str, value: Any, wrapper_maxsize: int | None = None) -> None:
        pickled_value = pickle.dumps(value, protocol=self._PROTO)
        with self._lock:
            cursor = self._conn.cursor()
            try:
                # Update last_accessed time on store as well
                current_time = time.time()
                cursor.execute(
                    f"""INSERT OR REPLACE INTO {self._TABLE_NAME}
                        (key, value, last_accessed) VALUES (?, ?, ?)""",
                    (key, pickled_value, current_time),
                )
                self._conn.commit()
            finally:
                cursor.close()

    def _evict_if_needed(self, wrapper_maxsize: int | None = None) -> None:
        # If this specific wrapper has no size limit, don't evict based on size.
        if wrapper_maxsize is None:
            return

        # Since _evict_if_needed is called *after* _store in the base class,
        # current_size reflects the size *after* the new item was added.
        current_size = self._get_current_size()

        # If current size is within the wrapper's limit, no eviction needed.
        if current_size <= wrapper_maxsize:
            return

        # Calculate how many items to evict (LRU based on last_accessed)
        limit = current_size - wrapper_maxsize
        if limit <= 0:
            return

        with self._lock:
            cursor = self._conn.cursor()
            try:
                # Find the keys of the `limit` least recently accessed items
                cursor.execute(
                    f"SELECT key FROM {self._TABLE_NAME} ORDER BY last_accessed ASC LIMIT ?",
                    (limit,),
                )
                keys_to_delete = [row[0] for row in cursor.fetchall()]

                if keys_to_delete:
                    # Delete the selected keys
                    # Use placeholders correctly for a list of values
                    placeholders = ",".join("?" * len(keys_to_delete))
                    cursor.execute(
                        f"DELETE FROM {self._TABLE_NAME} WHERE key IN ({placeholders})",
                        keys_to_delete,
                    )
                    self._conn.commit()
            finally:
                cursor.close()

    def _clear(self) -> None:
        """Clear all items from the cache table."""
        with self._lock:
            cursor = self._conn.cursor()
            try:
                cursor.execute(f"DELETE FROM {self._TABLE_NAME}")
                self._conn.commit()
            finally:
                cursor.close()

    def _get_current_size(self) -> int:
        """Return the number of rows in the cache table."""
        with self._lock:
            cursor = self._conn.cursor()
            try:
                cursor.execute(f"SELECT COUNT(*) FROM {self._TABLE_NAME}")
                count = cursor.fetchone()[0]
                return count
            finally:
                cursor.close()

    def __del__(self) -> None:
        if hasattr(self, "_conn") and self._conn:
            self._conn.close()



from __future__ import annotations
from typing import Any

import pickle
import time
from sqlalchemy import (
    Column,
    MetaData,
    String,
    Table,
    create_engine,
    select,
    text,
    func as sqlfunc,
    Float,
)
from sqlalchemy.engine.url import URL

from .base import _BaseCache, _MISSING


class SQLAlchemyCache(_BaseCache):
    """Persistent cache backed by any SQLAlchemy‑supported database."""

    def __init__(
        self,
        url: str | None = None,
        drivername: str | None = None,
        username: str | None = None,
        password: str | None = None,
        host: str | None = None,
        port: int | None = None,
        database: str | None = None,
        table_name: str = "picocache",
        echo: bool = False,
        **kw: Any,
    ) -> None:
        super().__init__(**kw)
        if url is None:
            if drivername is None:
                raise ValueError("Either *url* or *drivername* must be provided")

            url = str(
                URL.create(
                    drivername,
                    username=username,
                    password=password,
                    host=host,
                    port=port,
                    database=database,
                )
            )
        self._engine = create_engine(url, echo=echo, future=True)
        self._metadata = MetaData()
        self._table = Table(
            table_name,
            self._metadata,
            Column("key", String(64), primary_key=True),
            Column("value", String),
            Column("last_accessed", Float),
        )
        self._metadata.create_all(self._engine)

    def _lookup(self, key: str):
        with self._engine.begin() as conn:
            row = conn.execute(
                select(self._table.c.value).where(self._table.c.key == key)
            ).fetchone()
            if row is None:
                return _MISSING
            update_stmt = (
                self._table.update()
                .where(self._table.c.key == key)
                .values(last_accessed=time.time())
            )
            conn.execute(update_stmt)
            self._hits += 1
            return pickle.loads(bytes.fromhex(row.value))

    def _store(self, key: str, value: Any, wrapper_maxsize: int | None = None):
        pickled = pickle.dumps(value, protocol=self._PROTO).hex()
        current_time = time.time()
        with self._engine.begin() as conn:
            dialect = self._engine.dialect.name

            if dialect == "postgresql":
                stmt = self._table.insert().values(
                    key=key, value=pickled, last_accessed=current_time
                )
                stmt = stmt.on_conflict_do_update(
                    index_elements=["key"],
                    set_=dict(value=pickled, last_accessed=current_time),
                )
            elif dialect == "sqlite":
                stmt = text(
                    "INSERT OR REPLACE INTO {} (key, value, last_accessed) VALUES (:key, :value, :last_accessed)".format(
                        self._table.name
                    )
                )
            else:
                stmt = text(
                    "INSERT INTO {} (key, value, last_accessed) VALUES (:key, :value, :last_accessed) "
                    "ON CONFLICT(key) DO UPDATE SET value = :value, last_accessed = :last_accessed".format(
                        self._table.name
                    )
                )

            conn.execute(
                stmt, {"key": key, "value": pickled, "last_accessed": current_time}
            )

    def _evict_if_needed(self, wrapper_maxsize: int | None = None):
        if wrapper_maxsize is None:
            return

        current_size = self._get_current_size()

        if current_size <= wrapper_maxsize:
            return

        limit = current_size - wrapper_maxsize
        if limit <= 0:
            return

        with self._engine.begin() as conn:
            keys_to_delete_subquery = (
                select(self._table.c.key)
                .order_by(self._table.c.last_accessed.asc())
                .limit(limit)
                .subquery()
            )

            delete_stmt = self._table.delete().where(
                self._table.c.key.in_(select(keys_to_delete_subquery.c.key))
            )
            conn.execute(delete_stmt)

    def _clear(self) -> None:
        """Clear all items from the cache table."""
        with self._engine.begin() as conn:
            conn.execute(self._table.delete())

    def _get_current_size(self) -> int:
        """Return the number of rows in the cache table."""
        with self._engine.connect() as conn:
            count_query = select(sqlfunc.count()).select_from(self._table)
            result = conn.execute(count_query).scalar()
            return result if result is not None else 0
